/*
 * 
 * 
 * 
 * 
 * 
 * 
 */

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/etherdevice.h>
#include <linux/netdevice.h>

#include "w2lan.h"
#include "list.h"
#include "w2lan_output.h"


static struct list_head w2_head_list_TX;
static struct list_head w2_head_list_RX;
static struct list_head w2_head_list_convID;

static DEFINE_SPINLOCK(w2_lock_TX);
static DEFINE_SPINLOCK(w2_lock_RX);
static DEFINE_SPINLOCK(w2_lock_convID);



static struct kmem_cache *w2_slab;
static int count_convID_list = 0 ;

void del_foo(struct w2lan_data *w2,int type)
{
      spinlock_t *lock;

      if(type == W2_LIST_RX)	lock = &w2_lock_RX;	
      else lock = &w2_lock_TX;

      spin_lock(lock);
      list_del(&w2->list);
      spin_unlock(lock);
}


void put_foo(struct w2lan_data *w2)
{
      
      if (atomic_dec_and_test(&w2->refcnt))
      {
	if(w2->data!=NULL)	kfree(w2->data);
	kmem_cache_free(w2_slab, w2);
      }
}


static void w2_timer_expiry(unsigned long param)
{
	struct w2lan_data *w2 = (struct w2lan_data *)param;
	
	if(w2)
	{
		if(!w2->TX) 
		{    
		     del_foo(w2,W2_LIST_TX);
		     put_foo(w2);
		}
	        /*en caso de no recibir un Data despues de recibir un Announcio borramos el mensaje   */
		else if(w2->TX == 0 && w2->forward == 0 )
		{
		     del_foo(w2,W2_LIST_RX);
		     put_foo(w2);		  
		}
		else
		{
		    w2->TX = 1;
		    w2->forward = 0;
		    w2->timer.expires  = jiffies + w2lan_time;
		    add_timer(&w2->timer);
		}
	}

}

struct w2lan_data *get_data(unsigned char *convID,int type)
{
 
	  int count = 0;
 	  spinlock_t *lock;
	  struct w2lan_data *w2=NULL;
	  struct list_head *l_head;



          if(type == W2_LIST_RX)
	  {
		lock = &w2_lock_RX;
		l_head = &w2_head_list_RX;	
	  }
          else
	  { 
		lock = &w2_lock_TX;
		l_head = &w2_head_list_TX;
          }

	  spin_lock(lock);
	  list_for_each_entry(w2,l_head,list){  
	   
		if (compare_ether_addr(w2->convID,convID) == 0 ){
		count++;  
		atomic_inc(&w2->refcnt);
		break;
	      }
	  }
	 spin_unlock(lock);
	
	if(count==0) 	
		w2=NULL;
	return w2;
}

void w2lan_list_add_TX(struct w2lan_data *w2)
{
	spin_lock(&w2_lock_TX);
	list_add_tail(&w2->list,&w2_head_list_TX);
	spin_unlock(&w2_lock_TX);
}


void set_w2lan_data(struct w2lan_data *w2,int type)
{
	struct list_head *l_head;
	spinlock_t *lock;

	atomic_set(&w2->refcnt,1);
	init_timer(&w2->timer);
	w2->timer.expires  = jiffies + w2lan_time;
	w2->timer.function  = &w2_timer_expiry;
	w2->timer.data  = (unsigned long)w2;
	add_timer(&w2->timer);

	if(type == W2_LIST_RX)
	{
		w2->TX = 0;
		lock = &w2_lock_RX;
		l_head = &w2_head_list_RX;	
	}
	else
	{ 
		w2->TX = 1;
		lock = &w2_lock_TX;
		l_head = &w2_head_list_TX;
	}	  

	spin_lock(lock);
	list_add_tail(&w2->list,l_head);
	spin_unlock(lock); 

}


void w2lan_list_for_each_safe(struct list_head *l_head,spinlock_t *lock)
{

	struct list_head *p,*n;
	struct w2lan_data *w2=NULL;

retry:
	spin_lock(lock);
    	list_for_each_safe(p,n,l_head) {
		w2 = list_entry(p, struct w2lan_data, list);
 
		if(!del_timer(&w2->timer)){
			spin_unlock(lock);
			goto retry;
		  } 
		  list_del(&w2->list);
		  kmem_cache_free(w2_slab, w2);
	 }  
	 spin_unlock(lock);
}

int is_duplicate(struct w2lan_hdr_rcv *p)
{
	  struct w2lan_convID *convID_list;

	
	  spin_lock(&w2_lock_convID);

	  list_for_each_entry(convID_list,&w2_head_list_convID,list){  
	   
		if (compare_ether_addr(convID_list->convID,p->convID) == 0 ){
			printk("Duplicado \n ");
			spin_unlock(&w2_lock_convID);
			return 1;
		}      
	  }

	  convID_list = NULL;

		printk("count convID = %d \n ",count_convID_list);
		printk("list empty = %d \n ",list_empty(&w2_head_list_convID));

	  if(list_empty(&w2_head_list_convID) || count_convID_list < MaxConvID  )
	  {	
		convID_list = kmalloc (sizeof(struct w2lan_convID),GFP_ATOMIC);
		if(convID_list != NULL)
		{
			printk("nuevo duplicado guardado\n ");
			memcpy(convID_list->convID,p->convID,W2LAN_CONVID_LEN);
			memcpy(convID_list->orig_eth,p->orig_eth,W2LAN_ETH_HLEN);
			list_add_tail(&convID_list->list,&w2_head_list_convID);
			count_convID_list++;
		}
	  }
	  else
	  {
			printk("moviendo duplicado guardado\n ");
		convID_list = list_first_entry(&w2_head_list_convID,struct w2lan_convID,list);
			memcpy(convID_list->convID,p->convID,W2LAN_CONVID_LEN);
			memcpy(convID_list->orig_eth,p->orig_eth,W2LAN_ETH_HLEN);
		list_del(&convID_list->list);
		list_add_tail(&convID_list->list,&w2_head_list_convID);
	  }

	  spin_unlock(&w2_lock_convID);
	 return 0;
}

void purge_w2lan_list()
{
	  struct list_head *p,*n;
	  struct w2lan_convID *convID_list;

	  list_for_each_safe(p,n,&w2_head_list_convID){  
	   	convID_list = list_entry(p, struct w2lan_convID, list);
		list_del(&convID_list->list);
		kfree(convID_list);   
	  }

	  w2lan_list_for_each_safe(&w2_head_list_RX,&w2_lock_RX);
	  w2lan_list_for_each_safe(&w2_head_list_TX,&w2_lock_TX);
	  kmem_cache_destroy(w2_slab);
}



struct w2lan_data *w2lan_alloc()
{
	  struct w2lan_data *w2=NULL;
	  
	  w2 = kmem_cache_alloc(w2_slab,GFP_ATOMIC);
	  return w2;
}

static int  w2_slab_init(void)
{
        w2_slab = kmem_cache_create("w2lan_slab",
                                       sizeof(struct w2lan_data),
                                        0, 0, NULL);
        if (w2_slab == NULL)
                return -ENOMEM;
        return 0;
}

int list_init()
{
      int ret;
      ret = w2_slab_init();
      if(ret!=0) return ret;
      INIT_LIST_HEAD(&w2_head_list_TX);
      INIT_LIST_HEAD(&w2_head_list_RX);
      INIT_LIST_HEAD(&w2_head_list_convID);
      return ret;
}



