/*
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 */

#include <net/pkt_sched.h>

#include "w2lan.h"
#include "list.h"
#include "w2lan_output.h"

static unsigned char broadcast[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

void printHex1(char *mesg, unsigned char *p, int len)
{
        printk("%s\n", mesg);
        while(len--) {
	printk("%.2X ", *p);
	p++;
        }
        printk("\n");
}

struct sk_buff *w2_skb_create(struct w2lan_data *w2,unsigned char *daddr,int size,int type){
	
		  unsigned char *p;
		  struct sk_buff *skb;
		  struct ethhdr *eth; 


		  skb = alloc_skb(size + LL_ALLOCATED_SPACE(w2lan_dev),GFP_ATOMIC);
		
		  
		  if(skb == NULL)	
			return skb;		  
		  skb_reserve(skb,LL_RESERVED_SPACE(w2lan_dev));

		  skb_reset_network_header(skb);
		  p =(unsigned char*)skb_put(skb,size);

		  memcpy(p,w2->convID,W2LAN_CONVID_LEN);

		  if(type == SEND_DATA)
			  memcpy(&p[6],w2->data,w2->len);

		  if(type == SEND_ANNOUNCE)
			  memcpy(&p[6],w2->orig_eth,W2LAN_ETH_HLEN);		  

		  
		  eth = (struct ethhdr*)skb_push(skb,W2LAN_ETH_HLEN);
		  skb_reset_mac_header(skb);

		  if(type == SEND_ANNOUNCE && daddr == NULL)
		  {
			memcpy(eth->h_dest,broadcast,W2LAN_ETH_ALEN);
			memcpy(eth->h_source,w2lan_dev->dev_addr,W2LAN_ETH_ALEN);
			eth->h_proto = htons(W2LAN_A);
		  }
		  else if (type == SEND_REQUEST)
		  {
			memcpy(eth->h_dest,daddr,W2LAN_ETH_ALEN);
			memcpy(eth->h_source,w2lan_dev->dev_addr,W2LAN_ETH_ALEN);
			eth->h_proto = htons(W2LAN_R);
		  }
		  else
		  {
			memcpy(eth->h_dest,broadcast,W2LAN_ETH_ALEN);
			memcpy(eth->h_source,w2lan_dev->dev_addr,W2LAN_ETH_ALEN);
			eth->h_proto = htons(W2LAN_D);
		  }

		  skb->len = size + W2LAN_ETH_HLEN;
		  skb->dev = w2lan_dev;
		  return skb;
}



int w2lan_send_frame(struct w2lan_data *w2,unsigned char *daddr, int type)
{
	  int ret = -ENOMEM;
	  struct sk_buff *skb = NULL;

	  switch (type)
	  {
		case SEND_ANNOUNCE:
			skb = w2_skb_create(w2,NULL,W2LAN_ANNOUNCE_LEN,type);
			break;
		case SEND_REQUEST:
			skb = w2_skb_create(w2,daddr,W2LAN_REQUEST_LEN,type);
			break;
		case SEND_DATA:
			skb = w2_skb_create(w2,NULL,w2->len+W2LAN_CONVID_LEN,type);
			break;
		default:
		break;
	  }

	  if(skb != NULL)
		   ret = dev_queue_xmit(skb);
		  
	  return ret;
}

struct sk_buff *w2lan_generate_announce(struct sk_buff *skb)
{
	struct w2lan_data *w2=NULL;
	struct w2lan_hdr_rcv *w2_hdr=NULL;

 	w2 = w2lan_alloc();
	if(w2 == NULL )  
		    return NULL; 

	get_random_bytes(w2->convID,W2LAN_CONVID_LEN);

	memcpy(&w2->orig_eth,&skb->data[0],W2LAN_ETH_HLEN);

	w2->len = skb->len - W2LAN_ETH_HLEN;
	if(w2->len > 0)        
		w2->data = kmalloc(w2->len,GFP_ATOMIC);
        if(w2->data==NULL){
	   put_foo(w2);
	   kfree_skb(skb);
	   return NULL;
	}

	memcpy(&w2->data[0],&skb->data[W2LAN_ETH_HLEN],w2->len);
	kfree_skb(skb);
	skb = w2_skb_create(w2,NULL,W2LAN_ANNOUNCE_LEN,SEND_ANNOUNCE);
	set_w2lan_data(w2,W2_LIST_TX);
	w2_hdr = (struct w2lan_hdr_rcv *)eth_hdr(skb);
	is_duplicate(w2_hdr);
	
	return skb;
}

static int w2lan_enqueue(struct sk_buff *skb, struct Qdisc *sch)
{
	struct net_device *dev = qdisc_dev(sch);

	unsigned char *p;
	p = &skb->data[12];

	if((*p ^ w2_flag))
 	  skb = w2lan_generate_announce(skb);		

	if(skb == NULL)	{
		sch->qstats.drops++;
		return NET_XMIT_DROP;
	}
         if (sch->q.qlen < dev->tx_queue_len) {
		  __skb_queue_tail(&sch->q, skb);
                   sch->bstats.bytes += qdisc_pkt_len(skb);
                   sch->bstats.packets++;
                   return 0;
           }
	return NET_XMIT_DROP;
}




static int w2lan_qdisc_init(struct Qdisc *sch, struct nlattr *opt)
  {
          w2lan_dev = qdisc_dev(sch);

	  printk(KERN_DEBUG "W2LAN:dev[%s].\n ",w2lan_dev->name);
	  printk(KERN_DEBUG "W2LAN_ENQUEUE:sch id(%s).\n ",sch->ops->id);
          return 0;
  }

struct Qdisc_ops w2lan_qdisc_ops __read_mostly = {
  
      .id          =    "w2lan",
      .enqueue     =    w2lan_enqueue,
      .dequeue     =    qdisc_dequeue_head,
      .peek        =    qdisc_peek_head,
      .drop        =    qdisc_queue_drop,
      .init        =    w2lan_qdisc_init,
      .reset       =    qdisc_reset_queue,
      .owner  	   =    THIS_MODULE,
      
  };
EXPORT_SYMBOL(w2lan_qdisc_ops);
  
int init_w2lan_output(void)
{
      int ret;
      ret = register_qdisc(&w2lan_qdisc_ops);
      return ret;
}

void exit_w2lan_output(void)
{
      unregister_qdisc(&w2lan_qdisc_ops);
}
